from typing import Any, Optional

from langchain.chains.base import Chain

from lanarky.register import (
    STREAMING_CALLBACKS,
    STREAMING_JSON_CALLBACKS,
    WEBSOCKET_CALLBACKS,
)

from .agents import *  # noqa: F401, F403
from .base import (
    AsyncStreamingJSONResponseCallback,
    AsyncStreamingResponseCallback,
    AsyncWebsocketCallback,
)
from .llm import *  # noqa: F401, F403
from .retrieval_qa import *  # noqa: F401, F403

ERROR_MESSAGE = """Error! Chain type '{chain_type}' is not currently supported by '{callable_name}'.
Available chain types: {chain_types}

To use a custom chain type, you must register a new callback handler.
See the documentation for more details: https://lanarky.readthedocs.io/en/latest/advanced/custom_callbacks.html
"""


def get_streaming_callback(
    chain: Chain, override: Optional[str] = None, *args, **kwargs
) -> AsyncStreamingResponseCallback:
    """Get the streaming callback for the given chain type."""
    return _get_callback(
        chain,
        override,
        STREAMING_CALLBACKS,
        "AsyncStreamingResponseCallback",
        *args,
        **kwargs
    )


def get_websocket_callback(
    chain: Chain, override: Optional[str] = None, *args, **kwargs
) -> AsyncWebsocketCallback:
    """Get the websocket callback for the given chain type."""
    return _get_callback(
        chain, override, WEBSOCKET_CALLBACKS, "AsyncWebsocketCallback", *args, **kwargs
    )


def get_streaming_json_callback(
    chain: Chain, override: Optional[str] = None, *args, **kwargs
) -> AsyncStreamingJSONResponseCallback:
    """Get the streaming JSON callback for the given chain type."""
    return _get_callback(
        chain,
        override,
        STREAMING_JSON_CALLBACKS,
        "AsyncStreamingJSONResponseCallback",
        *args,
        **kwargs
    )


def _get_callback(
    chain: Chain,
    override: Optional[str],
    callback_registry: dict[str, Any],
    callable_name: str,
    *args,
    **kwargs
):
    """Base function for getting a callback from a registry.

    Args:
        chain: The chain to get the callback for.
        override: The name of the chain type to use instead of the chain's type.
        callback_registry: The registry to get the callback from.
        callable_name: The name of the callable to use in the error message.
        *args: Positional arguments to pass to the callback.
        **kwargs: Keyword arguments to pass to the callback.
    """
    chain_type = override or chain.__class__.__name__
    try:
        callback = callback_registry[chain_type]
        return callback(*args, **kwargs)
    except KeyError:
        raise KeyError(
            ERROR_MESSAGE.format(
                chain_type=chain_type,
                callable_name=callable_name,
                chain_types="\n".join(list(callback_registry.keys())),
            )
        )



from enum import Enum
from typing import Any, Union

from pydantic import BaseModel


class Sender(str, Enum):
    BOT = "bot"
    HUMAN = "human"


class Message(str, Enum):
    NULL = ""
    ERROR = "Sorry, something went wrong. Try again."


class MessageType(str, Enum):
    START = "start"
    STREAM = "stream"
    END = "end"
    ERROR = "error"
    INFO = "info"


class WebsocketResponse(BaseModel):
    sender: Sender
    message: Union[Message, str]
    message_type: MessageType

    class Config:
        use_enum_values = True


class StreamingJSONResponse(BaseModel):
    token: str = ""


class BaseRetrievalQAStreamingJSONResponse(StreamingJSONResponse):
    source_documents: list[dict[str, Any]]



from enum import IntEnum
from functools import lru_cache
from typing import Type

from fastapi import Depends, WebSocket, params
from langchain.chains.base import Chain
from pydantic import BaseModel, create_model

from lanarky.responses import StreamingResponse
from lanarky.websockets import WebsocketConnection


class StreamingMode(IntEnum):
    OFF = 0
    TEXT = 1
    JSON = 2


def create_langchain_dependency(langchain_object: Type[Chain]) -> params.Depends:
    """Creates a langchain object dependency."""

    @lru_cache(maxsize=1)
    def dependency() -> Chain:
        return langchain_object

    return Depends(dependency)


def create_request_from_langchain_dependency(
    langchain_dependency: params.Depends, name_prefix: str = ""
) -> Type[BaseModel]:
    langchain_object: Chain = langchain_dependency.dependency()
    model_name = name_prefix + "LangchainRequest"
    return create_model(
        model_name,
        **{key: (str, "") for key in langchain_object.input_keys},
    )


def create_response_model_from_langchain_dependency(
    langchain_dependency: params.Depends, name_prefix: str = ""
) -> Type[BaseModel]:
    """Creates a response model from a langchain dependency."""
    langchain_object: Chain = langchain_dependency.dependency()
    model_name = name_prefix + "LangchainResponse"
    return create_model(
        model_name,
        **{key: (str, "") for key in langchain_object.output_keys},
    )


def create_langchain_base_endpoint(
    endpoint_request: BaseModel,
    langchain_dependency: params.Depends,
    response_model: BaseModel,
):
    async def endpoint(
        request: endpoint_request,
        langchain_object: Chain = langchain_dependency,
    ) -> response_model:
        """Base chat endpoint."""
        return await langchain_object.acall(inputs=request.dict())

    return endpoint


def create_langchain_streaming_endpoint(
    endpoint_request: BaseModel, langchain_dependency: params.Depends
):
    async def endpoint(
        request: endpoint_request,
        langchain_object: Chain = langchain_dependency,
    ) -> StreamingResponse:
        """Streaming chat endpoint."""
        print(f"langchain_object: {langchain_object}")
        inputs = request.dict()
        print(f"inputs: {inputs}")
        return StreamingResponse.from_chain(
            langchain_object, inputs, media_type="text/event-stream"
        )

    return endpoint


def create_langchain_streaming_json_endpoint(
    endpoint_request: BaseModel, langchain_dependency: params.Depends
):
    async def endpoint(
        request: endpoint_request,
        langchain_object: Chain = langchain_dependency,
    ) -> StreamingResponse:
        """Streaming JSON chat endpoint."""
        return StreamingResponse.from_chain(
            langchain_object,
            request.dict(),
            as_json=True,
            media_type="text/event-stream",
        )

    return endpoint


def create_langchain_endpoint(
    endpoint_request: BaseModel,
    langchain_dependency: params.Depends,
    response_model: BaseModel,
    streaming_mode: StreamingMode,
):
    """Creates a Langchain endpoint."""
    if streaming_mode == StreamingMode.OFF:
        endpoint = create_langchain_base_endpoint(
            endpoint_request, langchain_dependency, response_model
        )
    elif streaming_mode == StreamingMode.TEXT:
        endpoint = create_langchain_streaming_endpoint(
            endpoint_request, langchain_dependency
        )
    elif streaming_mode == StreamingMode.JSON:
        endpoint = create_langchain_streaming_json_endpoint(
            endpoint_request, langchain_dependency
        )
    else:
        raise ValueError(f"Invalid streaming mode: {streaming_mode}")

    return endpoint


def create_langchain_websocket_endpoint(
    websocket: Type[WebSocket], langchain_dependency: params.Depends
):
    async def endpoint(
        websocket: websocket,
        langchain_object: Chain = langchain_dependency,
    ) -> None:
        """Websocket chat endpoint."""
        connection = WebsocketConnection.from_chain(
            chain=langchain_object, websocket=websocket
        )
        await connection.connect()

    return endpoint



from typing import Any

from lanarky.register import (
    register_streaming_callback,
    register_streaming_json_callback,
    register_websocket_callback,
)
from lanarky.schemas import StreamingJSONResponse

from .base import (
    AsyncStreamingJSONResponseCallback,
    AsyncStreamingResponseCallback,
    AsyncWebsocketCallback,
)

SUPPORTED_CHAINS = ["LLMChain", "ConversationChain"]


@register_streaming_callback(SUPPORTED_CHAINS)
class AsyncLLMChainStreamingCallback(AsyncStreamingResponseCallback):
    """AsyncStreamingResponseCallback handler for LLMChain."""

    async def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        """Run on new LLM token. Only available when streaming is enabled."""
        message = self._construct_message(token)
        await self.send(message)


@register_websocket_callback(SUPPORTED_CHAINS)
class AsyncLLMChainWebsocketCallback(AsyncWebsocketCallback):
    """AsyncWebsocketCallback handler for LLMChain."""

    async def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        """Run on new LLM token. Only available when streaming is enabled."""
        message = self._construct_message(token)
        await self.websocket.send_json(message)


@register_streaming_json_callback(SUPPORTED_CHAINS)
class AsyncLLMChainStreamingJSONCallback(AsyncStreamingJSONResponseCallback):
    """AsyncStreamingJSONResponseCallback handler for LLMChain."""

    async def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        """Run on new LLM token. Only available when streaming is enabled."""
        message = self._construct_message(StreamingJSONResponse(token=token))
        await self.send(message)



"""
Credits:

* `chat-langchain <https://github.com/hwchase17/chat-langchain>`_
* `langchain-chat-websockets <https://github.com/pors/langchain-chat-websockets>`_
"""
import logging
from typing import Any, Awaitable, Callable

from fastapi import WebSocket, WebSocketDisconnect
from langchain.chains.base import Chain
from pydantic import BaseModel, Field

from lanarky.callbacks import get_websocket_callback
from lanarky.schemas import Message, MessageType, Sender, WebsocketResponse

logger = logging.getLogger(__name__)


class BaseWebsocketConnection(BaseModel):
    websocket: WebSocket = Field(...)
    chain_executor: Callable[[str], Awaitable[Any]] = Field(...)

    class Config:
        arbitrary_types_allowed = True

    async def connect(self):
        await self.websocket.accept()
        while True:
            try:
                user_message = await self.websocket.receive_text()
                await self.websocket.send_json(
                    WebsocketResponse(
                        sender=Sender.HUMAN,
                        message=user_message,
                        message_type=MessageType.STREAM,
                    ).dict()
                )
                await self.websocket.send_json(
                    WebsocketResponse(
                        sender=Sender.BOT,
                        message=Message.NULL,
                        message_type=MessageType.START,
                    ).dict()
                )
                await self.chain_executor(user_message)
                await self.websocket.send_json(
                    WebsocketResponse(
                        sender=Sender.BOT,
                        message=Message.NULL,
                        message_type=MessageType.END,
                    ).dict()
                )
            except WebSocketDisconnect:
                logger.info("client disconnected.")
                break
            except Exception as e:
                logger.error(e)
                await self.websocket.send_json(
                    WebsocketResponse(
                        sender=Sender.BOT,
                        message=Message.ERROR,
                        message_type=MessageType.ERROR,
                    ).dict()
                )

    @staticmethod
    def _create_chain_executor(
        chain: Chain,
        websocket: WebSocket,
        response: WebsocketResponse,
        **callback_kwargs,
    ) -> Callable[[str], Awaitable[Any]]:
        raise NotImplementedError

    @classmethod
    def from_chain(
        cls,
        chain: Chain,
        websocket: WebSocket,
        callback_kwargs: dict[str, Any] = {},
    ) -> "BaseWebsocketConnection":
        chain_executor = cls._create_chain_executor(
            chain,
            websocket,
            WebsocketResponse(
                sender=Sender.BOT, message=Message.NULL, message_type=MessageType.STREAM
            ),
            **callback_kwargs,
        )

        return cls(
            chain_executor=chain_executor,
            websocket=websocket,
        )


class WebsocketConnection(BaseWebsocketConnection):
    """BaseLangchainStreamingResponse class wrapper for LLMChain instances."""

    @staticmethod
    def _create_chain_executor(
        chain: Chain,
        websocket: WebSocket,
        response: WebsocketResponse,
        **callback_kwargs,
    ) -> Callable[[str], Awaitable[Any]]:
        async def wrapper(user_message: str):
            return await chain.acall(
                inputs=user_message,
                callbacks=[
                    get_websocket_callback(
                        chain=chain,
                        websocket=websocket,
                        response=response,
                        **callback_kwargs,
                    )
                ],
            )

        return wrapper



from typing import Any

from lanarky.register import (
    register_streaming_callback,
    register_streaming_json_callback,
    register_websocket_callback,
)
from lanarky.schemas import BaseRetrievalQAStreamingJSONResponse

from .llm import (
    AsyncLLMChainStreamingCallback,
    AsyncLLMChainStreamingJSONCallback,
    AsyncLLMChainWebsocketCallback,
)

SUPPORTED_CHAINS = [
    "RetrievalQA",
    "ConversationRetrievalQA",
    "VectorDBQA",
    "QAWithSourcesChain",
    "VectorDBQAWithSourcesChain",
    "RetrievalQAWithSourcesChain",
    "ConversationalRetrievalChain",
]
SOURCE_DOCUMENT_TEMPLATE = """
page content: {page_content}
{document_metadata}
"""


@register_streaming_callback(SUPPORTED_CHAINS)
class AsyncBaseRetrievalQAStreamingCallback(AsyncLLMChainStreamingCallback):
    """AsyncStreamingResponseCallback handler for BaseRetrievalQA."""

    source_document_template: str = SOURCE_DOCUMENT_TEMPLATE

    async def on_chain_end(self, outputs: dict[str, Any], **kwargs: Any) -> None:
        """Run when chain ends running."""
        if "source_documents" in outputs:
            message = self._construct_message("\n\nSOURCE DOCUMENTS:\n")
            await self.send(message)
            for document in outputs["source_documents"]:
                document_metadata = "\n".join(
                    [f"{k}: {v}" for k, v in document.metadata.items()]
                )
                message = self._construct_message(
                    self.source_document_template.format(
                        page_content=document.page_content,
                        document_metadata=document_metadata,
                    )
                )
                await self.send(message)


@register_websocket_callback(SUPPORTED_CHAINS)
class AsyncBaseRetrievalQAWebsocketCallback(AsyncLLMChainWebsocketCallback):
    """AsyncWebsocketCallback handler for BaseRetrievalQA."""

    source_document_template: str = SOURCE_DOCUMENT_TEMPLATE

    async def on_chain_end(self, outputs: dict[str, Any], **kwargs: Any) -> None:
        """Run when chain ends running."""
        if "source_documents" in outputs:
            message = self._construct_message("\n\nSOURCE DOCUMENTS:\n")
            await self.websocket.send_json(message)
            for document in outputs["source_documents"]:
                document_metadata = "\n".join(
                    [f"{k}: {v}" for k, v in document.metadata.items()]
                )
                message = self._construct_message(
                    self.source_document_template.format(
                        page_content=document.page_content,
                        document_metadata=document_metadata,
                    )
                )
                await self.websocket.send_json(message)


@register_streaming_json_callback(SUPPORTED_CHAINS)
class AsyncBaseRetrievalQAStreamingJSONCallback(AsyncLLMChainStreamingJSONCallback):
    """AsyncStreamingJSONResponseCallback handler for BaseRetrievalQA."""

    async def on_chain_end(self, outputs: dict[str, Any], **kwargs: Any) -> None:
        """Run when chain ends running."""
        if "source_documents" in outputs:
            source_documents = [
                document.dict() for document in outputs["source_documents"]
            ]
            message = self._construct_message(
                BaseRetrievalQAStreamingJSONResponse(source_documents=source_documents)
            )
            await self.send(message)



import json
from abc import abstractmethod
from typing import Any

from fastapi import WebSocket
from langchain.callbacks.base import AsyncCallbackHandler
from pydantic import BaseModel, Field
from starlette.types import Message, Send

from lanarky.schemas import StreamingJSONResponse, WebsocketResponse


class AsyncLanarkyCallback(AsyncCallbackHandler, BaseModel):
    """Async Callback handler for FastAPI StreamingResponse."""

    @property
    def always_verbose(self) -> bool:
        """Whether to call verbose callbacks even if verbose is False."""
        return True

    class Config:
        arbitrary_types_allowed = True

    @abstractmethod
    def _construct_message(self, content: Any) -> Any:  # pragma: no cover
        """Construct a Message from a string."""
        pass


class AsyncStreamingResponseCallback(AsyncLanarkyCallback):
    """Async Callback handler for StreamingResponse."""

    send: Send = Field(...)

    def _construct_message(self, content: str) -> Message:
        """Construct a Message from a string."""
        return {
            "type": "http.response.body",
            "body": content.encode("utf-8"),
            "more_body": True,
        }


class AsyncWebsocketCallback(AsyncLanarkyCallback):
    """Async Callback handler for WebsocketConnection."""

    websocket: WebSocket = Field(...)
    response: WebsocketResponse = Field(...)

    def _construct_message(self, content: str) -> dict:
        """Construct a WebsocketResponse from a string."""
        return {**self.response.dict(), **{"message": content}}


class AsyncStreamingJSONResponseCallback(AsyncStreamingResponseCallback):
    """Async Callback handler for StreamingJSONResponse."""

    send: Send = Field(...)

    def _construct_message(self, content: StreamingJSONResponse) -> Message:
        """Construct a Message from a dictionary."""
        return {
            "type": "http.response.body",
            "body": json.dumps(
                content.dict(),
                ensure_ascii=False,
                allow_nan=False,
                indent=None,
                separators=(",", ":"),
            ).encode("utf-8"),
            "more_body": True,
        }



import random
import string
from typing import Any, Optional, Type

from fastapi.routing import APIRouter
from fastapi.websockets import WebSocket
from langchain.chains.base import Chain

from .utils import (
    StreamingMode,
    create_langchain_dependency,
    create_langchain_endpoint,
    create_langchain_websocket_endpoint,
    create_request_from_langchain_dependency,
    create_response_model_from_langchain_dependency,
)


class LangchainRouter(APIRouter):
    def __init__(
        self,
        *,
        langchain_url: Optional[str] = None,
        langchain_object: Optional[Type[Chain]] = None,
        langchain_endpoint_kwargs: Optional[dict[str, Any]] = None,
        streaming_mode: Optional[StreamingMode] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.langchain_url = langchain_url
        self.langchain_object = langchain_object
        self.langchain_endpoint_kwargs = langchain_endpoint_kwargs or {}
        self.streaming_mode = streaming_mode

        self.langchain_dependencies = []

        self.setup()

    def setup(self) -> None:
        """Sets up the Langchain router."""
        if self.langchain_url:
            self.add_langchain_api_route(
                self.langchain_url,
                self.langchain_object,
                self.streaming_mode,
                **self.langchain_endpoint_kwargs,
            )

    def add_langchain_api_route(
        self,
        url: str,
        langchain_object: Chain,
        streaming_mode: StreamingMode,
        methods: list[str] = ["POST"],
        **kwargs,
    ):
        """Adds a Langchain API route to the router."""
        langchain_dependency = create_langchain_dependency(langchain_object)

        name_prefix = "".join(
            random.choice(string.ascii_letters) for _ in range(5)
        ).title()
        endpoint_request = create_request_from_langchain_dependency(
            langchain_dependency, name_prefix
        )
        response_model = (
            create_response_model_from_langchain_dependency(
                langchain_dependency, name_prefix
            )
            if streaming_mode == StreamingMode.OFF
            else None
        )

        endpoint = create_langchain_endpoint(
            endpoint_request,
            langchain_dependency,
            response_model,
            streaming_mode,
        )

        self.add_api_route(
            url,
            endpoint,
            response_model=response_model,
            methods=methods,
            **kwargs,
        )

        self.langchain_dependencies.append(langchain_dependency)

    def add_langchain_api_websocket_route(self, url: str, langchain_object: Chain):
        """Adds a Langchain API websocket route to the router."""
        langchain_dependency = create_langchain_dependency(langchain_object)
        endpoint = create_langchain_websocket_endpoint(
            WebSocket,
            langchain_dependency,
        )

        self.add_api_websocket_route(url, endpoint)

        self.langchain_dependencies.append(langchain_dependency)
